import { fixBytes, mergeBytes } from './utils/bytes.mjs';
import { SdkError } from './errors/SdkError.mjs';

/**
 * An object that can serialize and deserialize a value to and from a `Uint8Array`.
 * It supports serializing looser types than it deserializes for convenience.
 * For example, a `bigint` serializer will always deserialize to a `bigint`
 * but can be used to serialize a `number`.
 *
 * @typeParam From - The type of the value to serialize.
 * @typeParam To - The type of the deserialized value. Defaults to `From`.
 *
 * @category Serializers
 */

function mapSerializer(serializer, unmap, map) {
  return {
    description: serializer.description,
    fixedSize: serializer.fixedSize,
    maxSize: serializer.maxSize,
    serialize: value => serializer.serialize(unmap(value)),
    deserialize: (buffer, offset = 0) => {
      const [value, length] = serializer.deserialize(buffer, offset);
      return map ? [map(value, buffer, offset), length] : [value, length];
    }
  };
}

/**
 * Creates a fixed-size serializer from a given serializer.
 *
 * @param serializer - The serializer to wrap into a fixed-size serializer.
 * @param fixedBytes - The fixed number of bytes to read.
 * @param description - A custom description for the serializer.
 *
 * @category Serializers
 */
function fixSerializer(serializer, fixedBytes, description) {
  return {
    description: description ?? `fixed(${fixedBytes}, ${serializer.description})`,
    fixedSize: fixedBytes,
    maxSize: fixedBytes,
    serialize: value => fixBytes(serializer.serialize(value), fixedBytes),
    deserialize: (buffer, offset = 0) => {
      // Slice the buffer to the fixed size.
      buffer = buffer.slice(offset, offset + fixedBytes);
      // Ensure we have enough bytes.
      if (buffer.length < fixedBytes) {
        throw new SdkError(`Fixed serializer expected ${fixedBytes} bytes, got ${buffer.length}.`);
      }
      // If the nested serializer is fixed-size, pad and truncate the buffer accordingly.
      if (serializer.fixedSize !== null) {
        buffer = fixBytes(buffer, serializer.fixedSize);
      }
      // Deserialize the value using the nested serializer.
      const [value] = serializer.deserialize(buffer, 0);
      return [value, offset + fixedBytes];
    }
  };
}

/**
 * Reverses the bytes of a fixed-size serializer.
 * @category Serializers
 */
function reverseSerializer(serializer) {
  if (serializer.fixedSize === null) {
    throw new SdkError('Cannot reverse a serializer of variable size.');
  }
  return {
    ...serializer,
    serialize: value => serializer.serialize(value).reverse(),
    deserialize: (bytes, offset = 0) => {
      const fixedSize = serializer.fixedSize;
      const newBytes = mergeBytes([bytes.slice(0, offset), bytes.slice(offset, offset + fixedSize).reverse(), bytes.slice(offset + fixedSize)]);
      return serializer.deserialize(newBytes, offset);
    }
  };
}

export { fixSerializer, mapSerializer, reverseSerializer };
//# sourceMappingURL=Serializer.mjs.map
